package {
    import Box2D.Collision.Shapes.b2PolygonDef;
    import Box2D.Collision.b2AABB;
    import Box2D.Common.Math.b2Vec2;
    import Box2D.Dynamics.Joints.b2RevoluteJointDef;
    import Box2D.Dynamics.b2Body;
    import Box2D.Dynamics.b2BodyDef;
    import Box2D.Dynamics.b2DebugDraw;
    import Box2D.Dynamics.b2World;
    
    import flash.display.Shape;
    import flash.display.Sprite;
    import flash.events.Event;
    import flash.events.MouseEvent;
    import flash.utils.getTimer;

    [SWF(width="800", height="600", frameRate="200")]

    public class Snake extends Sprite
    {
        // canvas dimensions 
        private var WIDTH  : int;
        private var HEIGHT : int;

        // snake "head"
        private var m_head  : b2Body;
        // physical world
        private var m_world : b2World;
        
        // head target point (== mouse pos)
        private var m_headTarget : b2Vec2 = new b2Vec2(-1, -1);

        // canvas with game graphics        
        private var m_gameCanvas  : Sprite = new Sprite();
        // canvas with debug graphics
        private var m_debugCanvas : Sprite = new Sprite();
        
        // last frame time in miliseconds
        private var m_lastFrameTime : int;
        
        // number of boxes along each axis
        private static const RANDOM_COUNT_PER_AXIS : int = 6;
                
        // snake parts count
        private static const PARTS_COUNT   : int    = 80;
        // snake part length
        private static const PART_LENGTH   : Number = 10.0;
        // snake part width      
        private static const PART_WIDTH    : Number = 10.0;
        // distance between snake parts
        // (less then part length to make parts overlap)
        private static const DISTANCE_BETWEEN_PARTS : Number = PART_LENGTH - 2.0;
        
        // angle to limit rotation of pair connected parts
        private static const PART_ANGLE_LIMIT : Number = 20.0 * Math.PI / 180.0;

        public function Snake()
        {
            // clip dimensions
            WIDTH  = stage.stageWidth;
            HEIGHT = stage.stageHeight;
            
            // add containers
            addChild(m_debugCanvas);
            addChild(m_gameCanvas);

            // world bounding box 
            var worldAABB : b2AABB = new b2AABB();
            worldAABB.lowerBound.Set(-WIDTH, -HEIGHT);
            worldAABB.upperBound.Set(2*WIDTH, 2*HEIGHT);            
            
            // create world. zero gravity + use "sleeping" for not moved objects 
            m_world = new b2World(worldAABB, b2Vec2.Make(0, 0), true);            

            // debug drawing
            var dbgDraw : b2DebugDraw = new b2DebugDraw();
            dbgDraw.m_sprite = m_debugCanvas;
            dbgDraw.m_drawScale     = 1.0;
            dbgDraw.m_fillAlpha     = 0.3;
            dbgDraw.m_lineThickness = 1.0;
            dbgDraw.m_drawFlags = b2DebugDraw.e_shapeBit | b2DebugDraw.e_jointBit;
            m_world.SetDebugDraw(dbgDraw);            

            // previos snake part, used for making joint
            var prevPart : b2Body = null;
            
            // for all parts
            for(var i : int = 0; i < PARTS_COUNT; i++)
            {
                // current part center (snake creates along +x axis)
                const xc : Number = 20 + i * DISTANCE_BETWEEN_PARTS;
                const yc : Number = 20;
                
                // snake part screen object
                var partDisplayObject : Shape = new Shape();
                // color from red to black
                partDisplayObject.graphics.beginFill((0xFF * (PARTS_COUNT - 1 - i) / (PARTS_COUNT - 1)) << 16);
                // rectangle
                partDisplayObject.graphics.drawRect(-PART_LENGTH/2, -PART_WIDTH/2, PART_LENGTH, PART_WIDTH);
                partDisplayObject.graphics.endFill();              

                // create physic object, save screen object in it                                 
                var curPart : b2Body = createBodyPart(xc, yc, PART_LENGTH, PART_WIDTH, partDisplayObject);
                
                // add screen object to the game canvas
                m_gameCanvas.addChild(partDisplayObject);
                
                // save head pointer
                if(0 == i)
                    m_head = curPart;
                
                // if previos part present, then link it with current 
                if(prevPart)
                {
                    // point joint
                    var jointDef : b2RevoluteJointDef = new b2RevoluteJointDef();
                    
                    // at the middle between parts
                    jointDef.Initialize(prevPart, curPart, b2Vec2.Make(xc - DISTANCE_BETWEEN_PARTS/2, yc));
                    // ignore collisions (but only between this to objects!)                     
                    jointDef.collideConnected = false;
                    // limit rotation
                    jointDef.enableLimit = true;
                    // for predefined angle
                    jointDef.lowerAngle = - PART_ANGLE_LIMIT;
                    jointDef.upperAngle = + PART_ANGLE_LIMIT;
                
                    // create and add joint to the world
                    m_world.CreateJoint(jointDef);
                }
                
                // current part become previos
                prevPart = curPart;
            }
            
            // add fps counter
            addChild(new FPSCounter());
            
            // random objects
            createRandomWorldObjects();

            // main "tick" 
            m_lastFrameTime = getTimer();
            addEventListener(Event.ENTER_FRAME, onEnterFrame, false, 0, true);

            // add listener for mouse move to whole clip area
            stage.addEventListener(MouseEvent.MOUSE_MOVE, onMouseMove, false, 0, true);
            // also click listener to switch debug - release
            stage.addEventListener(MouseEvent.CLICK, onMouseClick, false, 0, true);
            
            // at start gameCanvas visible (and debug is not)
            m_gameCanvas.visible = true;
            m_debugCanvas.visible = !m_gameCanvas.visible
        }
        
        // creates RANDOM_COUNT_PER_AXIS*RANDOM_COUNT_PER_AXIS little squares
        private function createRandomWorldObjects() : void
        {
            const dx : Number = WIDTH / RANDOM_COUNT_PER_AXIS;
            const dy : Number = HEIGHT / RANDOM_COUNT_PER_AXIS;
            
            const size : Number = 0.2 * (dx > dy ? dx : dy);
            
            var cy : Number = 0.5 * dy;
            for(var row : int = 0; row < RANDOM_COUNT_PER_AXIS; row++, cy += dy)
            {
                var cx : Number = 0.5 * dx;
                for(var col : int = 0; col < RANDOM_COUNT_PER_AXIS; col++, cx += dx)
                {
                    var randomElement : Shape = new Shape();
                    randomElement.graphics.beginFill(0xDDDDFF);
                    randomElement.graphics.drawRect(-size/2, -size/2, size, size);
                    randomElement.graphics.endFill();                    
                    createBodyPart(cx, cy, size, size, randomElement);
                    m_gameCanvas.addChild(randomElement);
                }
            }
        }

        // creates and registers rectangle object in the world
        // screen object saved in userData field of physical object
        private function createBodyPart(
            xc : Number, yc : Number,
            w : Number, h : Number,
            bodyDisplayObject : Shape) : b2Body
        {
            // object definition
            var bodyDef : b2BodyDef = new b2BodyDef();
            // world position
            bodyDef.position.Set(xc, yc);
            // linear "friction with world" 
            bodyDef.linearDamping  = 0.3;
            // angular "friction with world"
            bodyDef.angularDamping = 0.3;
            
            // shape definition
            var bodyShapeDef : b2PolygonDef = new b2PolygonDef();
            // as rectangle
            bodyShapeDef.SetAsBox(w/2, h/2);
            // density
            bodyShapeDef.density  = 1.0;
            // friction
            bodyShapeDef.friction = 0.3;
            
            // create object from definition
            var body : b2Body =  m_world.CreateBody(bodyDef);
            // create object shape from shape definition
            body.CreateShape(bodyShapeDef);
            // calculate mass, center of mass and inertia
            // based on shape
            body.SetMassFromShapes();
            
            // save screen object in userData field
            body.SetUserData(bodyDisplayObject);
            
            // return created body
            return body;
        }

        // each frame
        private function onEnterFrame(event : Event) : void
        {
            // dt - frame time in miliseconds
            const curFrameTime : int    = getTimer();
            const dt           : Number = 0.001 * (curFrameTime - m_lastFrameTime);
            m_lastFrameTime = curFrameTime;
            
            // if mouse was moved
            if(m_headTarget.x > 0)
            {
                // then add impulse to the snake head = 100 * vector from current position to the
                // mouse position
                const curHeadPos : b2Vec2 = m_head.GetWorldCenter();
                m_head.ApplyImpulse(
                    b2Vec2.Make(100.0 * (m_headTarget.x - curHeadPos.x), 100.0 * (m_headTarget.y - curHeadPos.y)),
                    curHeadPos);
            }
            
            // calculate physics
            // using 10 steps of physics for each frame  
            m_world.Step(dt, 10);

            // loop for all physic objects
            for(var body : b2Body = m_world.GetBodyList(); body; body = body.GetNext())
            {
                // if userData is not a sprite, then skip this object
                if(!(body.GetUserData() is Shape))
                    continue;

                // update screen object to be equal physical object                    
                body.GetUserData().x = body.GetPosition().x;
                body.GetUserData().y = body.GetPosition().y;
                body.GetUserData().rotation = body.GetAngle() * 180.0 / Math.PI;
            }
        }
        
        // on mouse move simply update headTarget
        private function onMouseMove(event : MouseEvent) : void
        {
            m_headTarget.Set(event.stageX, event.stageY);                        
        }

        // on mouse move click swap game/debug canvas visibilities
        private function onMouseClick(event : MouseEvent) : void
        {
            m_gameCanvas.visible = !m_gameCanvas.visible;
            m_debugCanvas.visible = !m_gameCanvas.visible;
        }
    }
}
